@@ -1,5 +1,5 @@
 /*
- * Copyright 2001 Sun Microsystems Inc.
+ * Copyright 2001, 2002, 2003 Sun Microsystems Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -22,15 +22,35 @@
 #include "htmlboxaccessible.h"
 #include "a11y/htmlviewaccessible.h"
 #include <libgtkhtml/layout/htmlboxtable.h>
+#include <libgtkhtml/layout/htmlboxtablecell.h>
 
 static void         html_box_table_accessible_class_init               (HtmlBoxTableAccessibleClass  *klass);
+static void         html_box_table_accessible_finalize                 (GObject          *object);
+static void	    html_box_table_accessible_real_initialize          (AtkObject        *object,
+                                               			        gpointer         data);
 
 static gint         html_box_table_accessible_get_n_children           (AtkObject        *obj);
 static AtkObject*   html_box_table_accessible_ref_child                (AtkObject        *obj,
                                                                         gint             i);
 
+static AtkObject*   find_cell (HtmlBoxTableAccessible *table,
+                               gint                   index);
+                                                                      
 static gpointer parent_class = NULL;
 
+struct _HtmlBoxTableAccessiblePrivate
+{
+	GList *extra_cells;
+};
+
+typedef struct _HtmlBoxTableAccessibleCellData HtmlBoxTableAccessibleCellData;
+
+struct _HtmlBoxTableAccessibleCellData
+{
+	gint index;
+	AtkObject *cell;
+};
+
 GType
 html_box_table_accessible_get_type (void)
 {
@@ -60,11 +80,50 @@ static void
 html_box_table_accessible_class_init (HtmlBoxTableAccessibleClass *klass)
 {
 	AtkObjectClass *class = ATK_OBJECT_CLASS (klass);
+	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
 
 	parent_class = g_type_class_peek_parent (klass);
 
+	gobject_class->finalize = html_box_table_accessible_finalize;
+
 	class->get_n_children = html_box_table_accessible_get_n_children;
 	class->ref_child = html_box_table_accessible_ref_child;
+	class->initialize = html_box_table_accessible_real_initialize;
+}
+
+static void
+html_box_table_accessible_real_initialize (AtkObject *object,
+                                           gpointer  data)
+{
+ 	HtmlBoxTableAccessible *table;
+
+	ATK_OBJECT_CLASS (parent_class)->initialize (object, data);
+
+	table = HTML_BOX_TABLE_ACCESSIBLE (object);
+	table->priv = g_new0 (HtmlBoxTableAccessiblePrivate, 1);
+}
+
+static void
+html_box_table_accessible_finalize (GObject *object)
+{
+	HtmlBoxTableAccessible *table;
+	GList *l;
+	HtmlBoxTableAccessibleCellData *cell_data;
+
+	table = HTML_BOX_TABLE_ACCESSIBLE (object);
+	if (table->priv) {
+		if (table->priv->extra_cells) {
+			for (l = table->priv->extra_cells; l; l = l->next) {
+				cell_data = (HtmlBoxTableAccessibleCellData *) (l->data);
+				g_object_unref (cell_data->cell);
+				g_free (l->data);
+			}
+			g_list_free (table->priv->extra_cells);
+		}
+		g_free (table->priv);
+	}
+
+	G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
 AtkObject*
@@ -122,7 +181,7 @@ html_box_table_accessible_ref_child (AtkObject *obj,
 {
 	AtkGObjectAccessible *atk_gobject;
 	HtmlBoxTable *box_table;
-	HtmlBoxTableRow *table_row;
+	HtmlBoxTableAccessibleCellData *cell_data;
 	HtmlBox *cell;
 	GObject *g_obj;
 	AtkObject *atk_child = NULL;
@@ -145,9 +204,37 @@ html_box_table_accessible_ref_child (AtkObject *obj,
 
 	index = g_slist_length (box_table->header_list) * box_table->cols + i;
 	cell = box_table->cells[index];
-	if (cell) {
+	if (!cell) {
+		atk_child = find_cell (HTML_BOX_TABLE_ACCESSIBLE (obj), index);
+		if (!atk_child) {
+			cell_data  = g_new (HtmlBoxTableAccessibleCellData, 1);
+			cell = html_box_table_cell_new ();
+			atk_child = atk_gobject_accessible_for_object (G_OBJECT (cell));
+			cell_data->cell = g_object_ref (atk_child);
+			atk_object_set_parent (atk_child, obj);
+			g_object_unref (cell);
+		}
+	} else {
 		atk_child = atk_gobject_accessible_for_object (G_OBJECT (cell));
-		g_object_ref (atk_child);
 	}
+	g_object_ref (atk_child);
 	return atk_child;
 }
+
+static AtkObject*
+find_cell (HtmlBoxTableAccessible *table,
+           gint                   index)
+{
+	GList *list, *l;
+	HtmlBoxTableAccessibleCellData *cell_data;
+
+	list = table->priv->extra_cells;
+	for (l = list; l; l = l->next) {
+		cell_data = (HtmlBoxTableAccessibleCellData *) (l->data);
+		if (cell_data->index == index)
+			return cell_data->cell;
+	}
+
+	return NULL;
+}
+
